# sdcat, Apache-2.0 license
# Filename: sdcat/cluster/cluster.py
# Description: Clustering using vision transformer features and HDBSCAN density-based clustering
import io
import multiprocessing
from collections import Counter
from importlib.util import find_spec

import pandas as pd
from pathlib import Path
import os
import json

import seaborn as sns
import numpy as np
from umap import UMAP
from hdbscan import HDBSCAN
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.preprocessing import MinMaxScaler
from sdcat.logger import info, warn, debug, err
from sdcat.cluster.utils import cluster_grid, crop_square_image, square_image, clean_bad_images
from sdcat.cluster.embedding import fetch_embedding, has_cached_embedding, compute_norm_embedding

if find_spec("multicore_tsne"):
    from multicore_tsne import MulticoreTSNE as TSNE
else:
    from sklearn.manifold import TSNE

if find_spec("cuml"):
    info('=======> USING GPU for HDBSCAN AND UMAP <=========')
    from cuml.cluster import HDBSCAN as cuHDBSCAN  # pylint: disable=E0611, E0401
    from cuml.manifold.umap import UMAP as cuUMAP

    have_gpu = True
else:
    have_gpu = False

sns.set_style("darkgrid", {"axes.facecolor": ".9"})
sns.set(rc={"figure.figsize": (12, 10)})


def top_majority(model_predictions, model_scores, threshold: float):
    """Find the top prediction"""

    # Only keep predictions with scores above the threshold
    data = [(pred, score) for pred, score in zip(model_predictions, model_scores) if float(score) >= threshold]

    if len(data) == 0:
        return None, None

    p, s = zip(*data)
    model_predictions = list(p)
    model_scores = list(s)

    # Count occurrences of each prediction in the top lists
    counter = Counter(model_predictions)

    majority_count = (len(model_predictions) // 2) + 1

    majority_predictions = [pred for pred, count in counter.items() if count >= majority_count]

    # If there are no majority predictions
    if len(majority_predictions) == 0:
        # Pick the prediction with the highest score if there is at least a .05 between the top two scores
        if len(model_predictions) > 1 and model_scores[0] - model_scores[1] >= 0.05:
            return model_predictions[0], model_scores[0]
        return None, None

    best_pred = majority_predictions[0]
    best_score = 0.0
    num_majority = 0
    # Sum all the scores for the majority predictions
    for pred, score in zip(model_predictions, model_scores):
        if pred in majority_predictions:
            best_score += float(score)
            num_majority += 1
    best_score /= num_majority

    return best_pred, best_score

def read_image(file_path: str) -> tuple[str, bytes]:
    with open(file_path, 'rb') as file:
        img = io.BytesIO(file.read()).getvalue()
        return file_path, img


def _run_hdbscan_assign(
        prefix: str,
        image_emb: np.ndarray,
        alpha: float,
        cluster_selection_epsilon: float,
        cluster_selection_method: str,
        min_similarity: float,
        min_cluster_size: int,
        min_samples: int,
        use_tsne: bool,
        ancillary_df: pd.DataFrame,
        out_path: Path) -> tuple:
    """
    Cluster the embeddings using HDBSCAN and reassign unclustered to the nearest exemplars.
    :param prefix:  A unique prefix to save artifacts from clustering
    :param image_emb:  The embeddings to cluster from the model
    :param alpha:  The alpha parameter for HDBSCAN
    :param cluster_selection_epsilon:  The epsilon parameter for HDBSCAN
    :param cluster_selection_method:  The method to use for cluster selection, 'leaf' or 'eom'
    :param min_similarity:  The minimum similarity score to use for clustering reassignment
    :param min_cluster_size:  The minimum number of samples in a cluster
    :param min_samples:   The number of samples in a neighborhood for a point
    :param use_tsne:  Whether to use t-SNE for dimensionality reduction
    :param ancillary_df:  (optional) Ancillary data to include in the clustering
    :param out_path:  The output path to save the clustering artifacts to
    :return: The average similarity score for each cluster, exemplar_df, cluster ids, cluster means, and coverage
    """
    info(f'Clustering using HDBSCAN using alpha {alpha} cluster_selection_epsilon {cluster_selection_epsilon} '
         f'min_samples {min_samples} use_tsne {use_tsne} ...')

    # Remove any existing cluster images in the output_path
    for c in out_path.parent.rglob(f'{prefix}_*cluster*.png'):
        c.unlink()

    # Add in any numerical ancillary data and replace NaNs with 0
    df = pd.DataFrame(image_emb)
    numerical = ancillary_df.select_dtypes(include=["float", "int"])
    if not numerical.empty:
        numerical = numerical.fillna(0)

        # Normalize the numerical data from 0 to 1 and add it to the dataframe
        numerical = (numerical - numerical.min()) / (numerical.max() - numerical.min())

        df = pd.merge(df, numerical, left_index=True, right_index=True, how='left')
        df = df.fillna(0)

    # Get the number of samples which is the number of rows in the dataframe - this is used mostly for calculating coverage
    num_samples = df.shape[0]

    # Perplexity must be less than the number of samples
    perplexity = min(30, num_samples - 1)

    # TSN-E does not work well when we have a few samples
    if num_samples > 100 and use_tsne:
        tsne = TSNE(n_components=2, perplexity=perplexity, metric="cosine", n_jobs=8, random_state=42, verbose=True)
        embedding = tsne.fit_transform(df.values)
    else:
        embedding = df.values
    x = MinMaxScaler().fit_transform(embedding)  # scale the embedding to 0-1

    # Cluster the embeddings using HDBSCAN
    if len(df) == 1:
        labels = np.array([0])
    else:
        if have_gpu:
            scan = cuHDBSCAN(
                metric='euclidean',  # 'precomputed' does not work with cuHDBSCAN
                allow_single_cluster=True,
                min_cluster_size=min_cluster_size,
                min_samples=min_samples,
                alpha=alpha,
                cluster_selection_epsilon=cluster_selection_epsilon,
                cluster_selection_method=cluster_selection_method)
            labels = scan.fit_predict(x)
        else:
            scan = HDBSCAN(
                    metric='l2',
                    allow_single_cluster=True,
                    min_cluster_size=min_cluster_size,
                    min_samples=min_samples,
                    alpha=alpha,
                    cluster_selection_epsilon=cluster_selection_epsilon,
                    cluster_selection_method=cluster_selection_method)
            labels = scan.fit_predict(x)

    # Get the unique clusters and sort them; -1 are unassigned clusters
    cluster_df = pd.DataFrame(labels, columns=['cluster'])
    unique_clusters = cluster_df['cluster'].unique().tolist()
    unique_clusters.sort()
    info(f"Number of clusters including unassigned -1 cluster: {len(unique_clusters)}")

    # If all the clusters are unassigned, then use all the samples as exemplars,
    # and assign them to the unknown cluster. If embedding is empty, this is also the case (failed to extract embeddings)
    if len(unique_clusters) == 1 and unique_clusters[0] == -1 or len(x) == 1:
        avg_sim_scores = []
        exemplar_df = pd.DataFrame()
        exemplar_df['cluster'] = len(x) * ['Unknown']
        exemplar_df['embedding'] = x.tolist()
        exemplar_df['crop_path'] = ancillary_df['crop_path'].tolist()
        clusters = []
        cluster_means = []
        coverage = 0.0
        return avg_sim_scores, exemplar_df, clusters, cluster_means, coverage

    cluster_df['score'] = scan.probabilities_
    # Get the index of the highest scores for each unique cluster sorted in increasing order
    # and use this as a representative image for the cluster
    max_scores = cluster_df.sort_values('cluster', ascending=True).groupby('cluster')['score'].idxmax()
    # Remove the last index which is the -1 cluster
    max_scores = max_scores[:-1]

    # Get the representative embeddings for the max scoring exemplars for each cluster and store them in a numpy array
    exemplar_emb = [image_emb[i] for i in max_scores]
    exemplar_emb = np.array(exemplar_emb)

    # Save the exemplar embeddings to a dataframe with some metadata
    exemplar_df = pd.DataFrame()
    exemplar_df['cluster'] = range(0, len(max_scores))  # Just use the index as the cluster id
    if ancillary_df is not None and 'image_path' in ancillary_df.columns:
        exemplar_df['crop_path'] = ancillary_df.iloc[max_scores]['crop_path'].tolist()
    exemplar_df['embedding'] = exemplar_emb.tolist()

    # Remove the last cluster which is the unknown cluster
    # Note that in the rare case where the coverage is 100%, the last cluster may not be the unknown cluster!
    # TODO: how to handle this case?
    max_clusters = len(unique_clusters) - 1
    num_before = len(exemplar_df)
    info(
        f"Removing {num_before - len(exemplar_df[exemplar_df['cluster'] != max_clusters])} samples from the unknown cluster")
    exemplar_df = exemplar_df[exemplar_df['cluster'] != max_clusters]
    num_after = len(exemplar_df)
    info(f"Number of samples after removing unknown cluster: {num_after}")

    # Reassign the unknowns to the closest cluster - this is only needed if the coverage is less than 1
    clustered = labels >= 0
    coverage = np.sum(clustered) / num_samples
    if coverage < 1.0:
        # Reassign based on the closest distance to exemplar
        for i, label in enumerate(labels):
            if label == -1:
                similarity_scores = cosine_similarity(image_emb[i].reshape(1, -1), exemplar_emb)
                closest_match_index = np.argmax(similarity_scores)
                # Only reassign if the similarity score is above the threshold
                if similarity_scores[0][closest_match_index] >= min_similarity:
                    labels[i] = closest_match_index

    clusters = [[] for _ in range(len(unique_clusters))]

    # Assign indices to the clusters
    for i in range(0, len(labels)):
        clusters[labels[i]].append(i)

    # Compute the average similarity score for each cluster
    avg_sim_scores = []
    for i, c in enumerate(clusters):
        debug(f'Computing similarity for cluster {i} with {len(c)} samples')
        if len(c) == 0:
            avg_sim_scores.append(0)
            continue
        cosine_sim_matrix = cosine_similarity(image_emb[c])
        avg_sim_scores.append(np.mean(cosine_sim_matrix))

    # Compute the cluster means
    cluster_means = []
    for c in clusters:
        cluster_means.append(np.mean(image_emb[c], axis=0))

    hdbscan_params = {
        "min_cluster_size": min_cluster_size,
        "min_samples": min_samples,
        "cluster_selection_method": "leaf",
        "metric": "precomputed",
        "alpha": alpha,
        "cluster_selection_epsilon": cluster_selection_epsilon
    }
    clustered = labels >= 0
    coverage = np.sum(clustered) / num_samples
    params = {
        "coverage": coverage,
        "num_clusters": len(unique_clusters),
        "hdbscan_params": hdbscan_params,
    }
    info(f"Parameters {params}")

    # Cannot use init='spectral' when n_components is >= num_samples - default to 'random' instead
    n_components = min(2, num_samples)
    if n_components >= num_samples:
        init = 'random'
    else:
        init = 'spectral'

    # Reduce the dimensionality of the embeddings using UMAP to 2 dimensions to visualize the clusters
    n_neighbors = min(15, df.values.shape[0] - 1)
    info(f'Using {n_neighbors} neighbors for dimensional reduction')
    if n_neighbors < 2:
        warn('Using PCA instead of UMAP')
        from sklearn.decomposition import PCA
        pca = PCA(n_components=2)
        xx = pca.fit_transform(df.values)
    else:
        if have_gpu:
            xx = cuUMAP(init=init,
                        n_components=2,
                        n_neighbors=n_neighbors,
                        min_dist=0.1,
                        metric='euclidean').fit_transform(df.values)
        else:
            xx = UMAP(init=init,
                      n_components=2,
                      n_neighbors=n_neighbors,
                      metric='cosine',
                      low_memory=True).fit_transform(df.values)

    df = pd.DataFrame({'x': xx[clustered, 0], 'y': xx[clustered, 1], 'labels': labels[clustered]})
    p = sns.jointplot(data=df, x='x', y='y', hue='labels')
    p.savefig(f"{out_path}/{prefix}_summary.png")
    info(f"Saved {out_path}/{prefix}_summary.png")

    with open(f'{out_path}/{prefix}_summary.json', 'w') as f:
        json.dump(params, f)

    return avg_sim_scores, exemplar_df, clusters, cluster_means, coverage


def cluster_vits(
        prefix: str,
        model: str,
        df_dets: pd.DataFrame,
        output_path: Path,
        alpha: float,
        cluster_selection_epsilon: float,
        cluster_selection_method: str,
        min_similarity: float,
        min_cluster_size: int,
        min_samples: int,
        device: str = "cpu",
        use_vits: bool = False,
        use_tsne: bool = False,
        skip_visualization: bool = False,
        remove_bad_images: bool = False,
        roi: bool = False) -> pd.DataFrame:
    """  Cluster the crops using the VITS embeddings.
    :param prefix:  A unique prefix to save artifacts from clustering
    :param model: The model to use for clustering
    :param df_dets: The dataframe with the detections
    :param output_path: The output path to save the clustering artifacts to
    :param roi:  Whether the detections are already cropped to the ROI
    :param cluster_selection_epsilon: The epsilon parameter for HDBSCAN
    :param cluster_selection_method: The method to use for cluster selection, 'leaf' or 'eom'
    :param alpha: The alpha parameter for HDBSCAN
    :param min_similarity: The minimum similarity score to use for -1 cluster reassignment
    :param min_cluster_size: The minimum number of samples in a cluster
    :param min_samples:The number of samples in a neighborhood for a point
    :param device: The device to use for clustering, 'cpu' or 'cuda'
    :param use_vits: Set to using the predictions from the vits cluster model
    :param skip_visualization: Whether to skip the visualization of the clusters
    :param remove_bad_images: Whether to remove bad images from the clustering
    :param use_tsne: Whether to use t-SNE for dimensionality reduction
    :return:  a dataframe with the assigned cluster indexes, or -1 for non-assigned."""

    # If there are no detections, return an empty dataframe
    if df_dets.empty:
        warn('No detections found in {detections} ')
        return pd.DataFrame()

    # Count how many files exists
    num_crop = sum([os.path.exists(filename) for filename in df_dets['crop_path']])

    # Skip cropping if all the crops are already done
    if num_crop != len(df_dets):
        num_processes = min(multiprocessing.cpu_count(), len(df_dets))
        if roi is True:
            info(f'ROI crops already exist')
        else:
            # Crop and squaring the images in parallel using multiprocessing to speed up the processing
            info(f'Cropping {len(df_dets)} detections in parallel using {num_processes} processes...')
            with multiprocessing.Pool(num_processes) as pool:
                args = [(row, 224) for index, row in df_dets.iterrows()]
                pool.starmap(crop_square_image, args)

    if remove_bad_images:
        # Remove any detections that are in any corner of the image
        size_before = len(df_dets)
        df = clean_bad_images(df_dets)
        size_after = len(df)
        info(f'Removed {size_before - size_after} detections that were dark or blurry')

    # Drop any rows with crop_path that have files that don't exist - sometimes the crops fail
    df_dets = df_dets[df_dets['crop_path'].apply(lambda x: os.path.exists(x))]
    df_dets = df_dets.copy()

    # Get the list of images to crop
    images = df_dets['crop_path'].tolist()

    # Count how many files have the .npy extension
    num_cached = sum([has_cached_embedding(model, filename) for filename in images])

    # Skip the embedding extraction if all the embeddings are cached
    if num_cached != len(images):
        debug(f'Extracted embeddings from {len(images)} images using model {model}...')
        compute_norm_embedding(model, images, device)

    # Fetch the cached embeddings
    debug('Fetching embeddings ...')
    image_emb = []
    for filename in images:
        emb, label, score = fetch_embedding(model, filename)
        if len(emb) == 0:
            # If the embeddings are zero, then the extraction failed; add a zero array
            image_emb.append(np.zeros(384, dtype=np.float32))
        else:
            image_emb.append(emb)
        if use_vits:
            df_dets.loc[df_dets['crop_path'] == filename, 'class'] = label[0]
            df_dets.loc[df_dets['crop_path'] == filename, 'score'] = score[0]

    # If the embeddings are zero, then the extraction failed
    num_failed = [i for i, e in enumerate(image_emb) if np.all(e == 0)]
    if len(num_failed) == len(images):
        warn('Failed to extract embeddings from all images')
        return pd.DataFrame()

    image_emb = np.array(image_emb)

    if not (output_path / prefix).exists():
        (output_path / prefix).mkdir(parents=True)

    # Remove everything except ancillary data to include in clustering
    columns = ['x', 'y', 'xx', 'xy', 'w', 'h', 'image_width', 'image_height', 'cluster_id', 'cluster', 'score',
               'class', 'image_path', 'crop_path']
    # Check if the columns exist in the dataframe
    if all(col in df_dets.columns for col in columns):
        ancillary_df = df_dets.drop(
            columns=['x', 'y', 'xx', 'xy', 'w', 'h', 'image_width', 'image_height', 'cluster_id', 'cluster', 'score',
                     'class', 'image_path', 'crop_path'])
    else:
        ancillary_df = df_dets

    # Cluster the images
    cluster_sim, exemplar_df, unique_clusters, cluster_means, coverage = _run_hdbscan_assign(prefix,
                                                                                             image_emb,
                                                                                             alpha,
                                                                                             cluster_selection_epsilon,
                                                                                             cluster_selection_method,
                                                                                             min_similarity,
                                                                                             min_cluster_size,
                                                                                             min_samples,
                                                                                             use_tsne,
                                                                                             ancillary_df,
                                                                                             output_path / prefix)

    # Get the average similarity across all clusters
    avg_similarity = np.mean(cluster_sim)

    info(f'Average similarity: {avg_similarity:.2f} min {min_similarity:.2f}  ')

    if len(unique_clusters) == 0:
        warn('No clusters found')
        # Save the exemplar embeddings with the model type
        exemplar_df['model'] = model
        exemplar_df.to_csv(output_path / f'{prefix}_exemplars.csv', index=False)
        return None

    info(f'Found {len(unique_clusters)} clusters with an average similarity of {avg_similarity:.2f} ')

    # Assign the cluster ids to the dataframe
    for cluster_id, cluster in enumerate(unique_clusters):
        for idx in cluster:
            debug(f'Adding {images[idx]} to cluster id {cluster_id} ')
            df_dets.loc[df_dets['crop_path'] == images[idx], 'cluster'] = cluster_id

    # For each cluster  let's create a grid of the images to check the quality of the clustering results
    num_processes = min(multiprocessing.cpu_count(), len(unique_clusters))
    info(f'Using {num_processes} processes to visualize the {len(unique_clusters)} clusters')

    if num_processes == 0:
        err(f'No processes available to visualize the clusters')
        return None

    if not skip_visualization:
        # Use a pool of processes to speed up the visualization of the clusters
        with multiprocessing.Pool(num_processes) as pool:
            args = [(prefix,  # prefix
                     cluster_sim[cluster_id],  # average similarity for the cluster
                     cluster_id,  # cluster id
                     unique_clusters[cluster_id],  # cluster indices
                     4 if len(unique_clusters[cluster_id]) < 50 else 8,  # grid size; larger clusters get larger grids
                     [images[idx] for idx in unique_clusters[cluster_id]],  # images in the cluster
                     output_path / prefix) for cluster_id in
                    range(0, len(unique_clusters))]
            pool.starmap(cluster_grid, args)

    # Save the exemplar embeddings with the model type
    exemplar_df['model'] = model
    exemplar_df.to_csv(output_path / f'{prefix}_exemplars.csv', index=False)

    info(f"Number of images {len(images)}")
    info(f"Number of clusters {len(unique_clusters)}")
    info(f"Coverage {coverage:.2f}")

    return df_dets



# sdcat, Apache-2.0 license
# Filename: sdcat/cluster/utils.py
# Description: Miscellaneous utility functions for cropping, clustering, and saving detections
import os

import cv2
import numpy as np
import pandas as pd
from PIL import Image
from cleanvision import Imagelab
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import ImageGrid
from pathlib import Path
from sdcat.logger import info, debug, warn, exception


def cluster_grid(prefix: str, cluster_sim: float, cluster_id: int, cluster_size: int, nb_images_display: int,
                 images: list, output_path: Path):
    """
    Cluster visualization; create a grid of images
    :param cluster_sim: Cluster similarity
    :param cluster_size: Size of the cluster
    :param cluster_id: Cluster ID
    :param images: list of images
    :param output_path: output path to save the visualization to
    :return:
    """
    debug(f'Cluster number {cluster_id} size {len(cluster_size)} similarity {cluster_sim}\n')

    # Plot a grid for each group of images nb_images_display at a time (e.g. 8x8)
    for i in range(0, len(images), nb_images_display * nb_images_display):
        fig = plt.figure(figsize=(10., 10.))
        grid = ImageGrid(fig, 111,  # similar to subplot(111)
                         nrows_ncols=(nb_images_display, nb_images_display),
                         # creates nb_images_display x nb_images_display grid of axes
                         axes_pad=0.025,
                         share_all=True,
                         cbar_pad=0.025)
        images_display = images[i:i + nb_images_display * nb_images_display]
        page = i // (nb_images_display * nb_images_display)

        # If we have more than 3 pages, then only display the first 3 pages
        # There can be a large number of pages for detections in common classes
        if page > 3:
            break

        total_pages = len(images) // (nb_images_display * nb_images_display)
        # debug(f"{i} Image filename:", images[j])
        for j, image in enumerate(images_display):
            try:
                image_square = Image.open(image)
                grid[j].imshow(image_square)
            except Exception as e:
                exception(f'Error opening {image} {e}')
                continue

            grid[j].axis('off')
            # If the verified is in the image name, then add a label to the image in the top center corner
            if 'verified' in image:
                n = Path(image)
                title = f"{n.stem.split('_')[0]}"
                grid[j].text(30, 10, title, fontsize=8, color='white', ha='center', va='center')
            # clear the x and y-axis
            grid[j].set_xticklabels([])

        # Add a title to the figure
        if total_pages > 1:
            fig.suptitle(
                f"{prefix} Cluster {cluster_id}, Size: {len(cluster_size)}, Similarity: {cluster_sim:.2f}, Page: {page} of {total_pages}",
                fontsize=16)
        else:
            fig.suptitle(f"{prefix} Cluster {cluster_id}, Size: {len(cluster_size)}, Similarity: {cluster_sim:.2f}",
                         fontsize=16)

        # Set the background color of the grid to white
        fig.set_facecolor('white')

        # Write the figure to a file
        out = output_path / f'{prefix}_cluster_{cluster_id}_p{page}.png'
        debug(f'Writing {out}')
        fig.savefig(out.as_posix())
        plt.close(fig)



def square_image(row, square_dim: int):
    """
    Squares an image to the model dimension, filling it with black bars if necessary
    :param row:
    :param square_dim: dimension of the square image
    :return:
    """
    try:
        if not Path(row.image_path).exists():
            warn(f'Skipping {row.crop_path} because the image {row.image_path} does not exist')
            return

        if Path(row.crop_path).exists():  # If the crop already exists, skip it
            return

        # Determine the size of the new square
        max_side = max(row.image_width, row.image_height)

        # Create a new square image with a black background
        new_image = Image.new('RGB', (max_side, max_side), (0, 0, 0))

        img = Image.open(row.image_path)

        # Paste the original image onto the center of the new image
        new_image.paste(img, ((max_side - row.image_width) // 2, (max_side - row.image_height) // 2))

        # Resize the image to square_dim x square_dim
        img = img.resize((square_dim, square_dim), Image.LANCZOS)

        # Save the image
        img.save(row.crop_path)
        img.close()
    except Exception as e:
        exception(f'Error cropping {row.image_path} {e}')
        raise e


def crop_square_image(row, square_dim: int):
    """
    Crop the image to a square padding the shortest dimension, then resize it to square_dim x square_dim
    This also adjusts the crop to make sure the crop is fully in the frame, otherwise the crop that
    exceeds the frame is filled with black bars - these produce clusters of "edge" objects instead
    of the detection
    :param row:
    :param square_dim: dimension of the square image
    :return:
    """
    try:

        if not Path(row.image_path).exists():
            warn(f'Skipping {row.crop_path} because the image {row.image_path} does not exist')
            return

        if Path(row.crop_path).exists():  # If the crop already exists, skip it
            return

        x1 = int(row.image_width * row.x)
        y1 = int(row.image_height * row.y)
        x2 = int(row.image_width * row.xx)
        y2 = int(row.image_height * row.xy)
        width = x2 - x1
        height = y2 - y1
        shorter_side = min(height, width)
        longer_side = max(height, width)
        delta = abs(longer_side - shorter_side)

        # Divide the difference by 2 to determine how much padding is needed on each side
        padding = delta // 2

        # Add the padding to the shorter side of the image
        if width == shorter_side:
            x1 -= padding
            x2 += padding
        else:
            y1 -= padding
            y2 += padding

        # Make sure that the coordinates don't go outside the image
        # If they do, adjust by the overflow amount
        if y1 < 0:
            y1 = 0
            y2 += abs(y1)
            if y2 > row.image_height:
                y2 = row.image_height
        elif y2 > row.image_height:
            y2 = row.image_height
            y1 -= abs(y2 - row.image_height)
            if y1 < 0:
                y1 = 0
        if x1 < 0:
            x1 = 0
            x2 += abs(x1)
            if x2 > row.image_width:
                x2 = row.image_width
        elif x2 > row.image_width:
            x2 = row.image_width
            x1 -= abs(x2 - row.image_width)
            if x1 < 0:
                x1 = 0

        # Crop the image
        img = Image.open(row.image_path)
        img = img.crop((x1, y1, x2, y2))

        # Resize the image to square_dim x square_dim
        img = img.resize((square_dim, square_dim), Image.LANCZOS)

        # Save the image
        img.save(row.crop_path)
        img.close()

    except Exception as e:
        exception(f'Error cropping {row.image_path} {e}')
        raise e


def rescale(img: np.ndarray, scale_percent: int = 75) -> np.ndarray:
    """
    Rescale an image
    :param img: Image to rescale
    :param scale_percent: Scale percentage
    :return: Rescaled image
    """
    width, height = img.shape[:2]
    width = int(width * scale_percent / 100)
    height = int(height * scale_percent / 100)
    dim = (height, width)

    # Resize the image to the new dimensions exactly
    img_rescaled = cv2.resize(img, dim, interpolation=cv2.INTER_AREA)
    return img_rescaled


def clean_bad_images(df: pd.DataFrame) -> pd.DataFrame:
    """Remove dark or blurry images from the dataframe"""
    crop_path = df['crop_path'].tolist()
    data_path = Path(crop_path[0]).parent # just take the first image to get the data path
    imagelab = Imagelab(data_path=data_path)
    imagelab.find_issues()
    # Columns to check for issues
    issue_columns = ["is_dark_issue", "is_blurry_issue", "is_near_duplicates_issue", "is_low_information_issue", "is_exact_duplicates_issue"]
    bad_images  = imagelab.issues[imagelab.issues[issue_columns].any(axis=1)].index
    num_removed = len(bad_images)
    debug(f"Removing {num_removed} dark, blurry or duplicate images in {crop_path}")
    for img in bad_images:
        os.remove(img)
    # Remove the bad images from the dataframe
    df = df[~df['crop_path'].isin(bad_images)]
    info(f"Removed {num_removed} dark or blurry images in {crop_path}")
    return df



# sdcat, Apache-2.0 license
# Filename: sdcat/cluster/embedding.py
# Description:  Miscellaneous functions for computing VIT embeddings and caching them.

import os
import multiprocessing
from PIL import Image
from numpy import save, load
import numpy as np
from sahi.utils.torch import torch
import cv2
from transformers import AutoModelForImageClassification, AutoImageProcessor
import torch.nn.functional as F

from sdcat.logger import info, err


class ViTWrapper:
    DEFAULT_MODEL_NAME = "google/vit-base-patch16-224"

    def __init__(self, device: str = "cpu", batch_size: int = 32, model_name: str = DEFAULT_MODEL_NAME):
        self.batch_size = batch_size
        self.name = model_name
        self.model = AutoModelForImageClassification.from_pretrained(model_name)
        self.processor = AutoImageProcessor.from_pretrained(model_name)

        # Load the model and processor
        if 'cuda' in device and torch.cuda.is_available():
            device_num = int(device.split(":")[-1])
            info(f"Using GPU device {device_num}")
            torch.cuda.set_device(device_num)
            self.device = "cuda"
            self.model.to("cuda")
        else:
            self.device = "cpu"

    @property
    def model_name(self) -> str:
        return self.name

    @property
    def vector_dimensions(self) -> int:
        return self.model.config.hidden_size

    def process_images(self, image_paths: list):
        info(f"Processing {len(image_paths)} images with {self.model_name}")

        images = [Image.open(image_path).convert("RGB") for image_path in image_paths]
        inputs = self.processor(images=images, return_tensors="pt").to(self.device)

        with torch.no_grad():
            outputs = self.model(**inputs)
            logits = outputs.logits
            embeddings = self.model.base_model(**inputs)
            batch_embeddings = embeddings.last_hidden_state[:, 0, :].cpu().numpy()
            # predicted_class_idx = logits.argmax(-1).cpu().numpy()
            # predicted_classes = [self.model.config.id2label[class_idx] for class_idx in predicted_class_idx]
            # predicted_scores = F.softmax(logits, dim=-1).cpu().numpy()
            # Get the top 3 classes and scores
            top_scores, top_classes = torch.topk(logits, 3)
            top_classes = top_classes.cpu().numpy()
            top_scores = F.softmax(top_scores, dim=-1).cpu().numpy()
            predicted_classes = [",".join([self.model.config.id2label[class_idx] for class_idx in class_list]) for class_list in top_classes]
            predicted_scores = [",".join([str(score) for score in score_list]) for score_list in top_scores]

        return batch_embeddings, predicted_classes, predicted_scores

def cache_embedding(embedding, pred, score, model_name: str, filename: str):
    model_machine_friendly_name = model_name.replace("/", "_")
    # save embeddings numpy array as npy file and the predictions as a txt file
    save(f'{filename}_{model_machine_friendly_name}.npy', embedding)
    with open(f'{filename}_{model_machine_friendly_name}_pred.txt', 'w') as f:
        predictions = pred.split(',')
        scores = score.split(',')
        for pred, score in zip(predictions, scores):
            f.write(f'{pred.strip()},{score.strip()}\n')


def fetch_embedding(model_name: str, filename: str):
    model_machine_friendly_name = model_name.replace("/", "_")
    # if the npy file exists, return it
    emb = []
    label = []
    score = []
    if os.path.exists(f'{filename}_{model_machine_friendly_name}.npy'):
        emb = load(f'{filename}_{model_machine_friendly_name}.npy')
    else:
        info(f'No embedding found for {filename}')
    if os.path.exists(f'{filename}_{model_machine_friendly_name}_pred.txt'):
        with open(f'{filename}_{model_machine_friendly_name}_pred.txt', 'r') as f:
            lines = f.readlines()
            for line in lines:
                line = line.strip().split(',')
                label.append(line[0])
                score.append(float(line[1]))
    else:
        info(f'No prediction found for {filename}')
    return emb, label, score


def has_cached_embedding(model_name: str, filename: str) -> int:
    """
    Check if the given filename has a cached embedding
    :param model_name: Name of the model
    :param filename: Name of the file
    :return: 1 if the image has a cached embedding, otherwise 0
    """
    model_machine_friendly_name = model_name.replace("/", "_")
    if os.path.exists(f'{filename}_{model_machine_friendly_name}.npy') and \
            os.path.exists(f'{filename}_{model_machine_friendly_name}_pred.txt'):
        return 1
    return 0


def encode_image(filename):
    img = Image.open(filename)
    keep = img.copy()
    img.close()
    return keep


def compute_embedding_vits(vit:ViTWrapper, images: list):
    """
    Compute the embedding for the given images using the given model
    :param vitwrapper: Wrapper for the ViT model
    :param images: List of image filenames
    :param model_name: Name of the model (i.e. google/vit-base-patch16-224, dinov2_vits16, etc.)
    :param device: Device to use for the computation (cpu or cuda:0, cuda:1, etc.)
    """
    batch_size = 32
    model_name = vit.model_name

    # Batch process the images
    batches = [images[i:i + batch_size] for i in range(0, len(images), batch_size)]
    for batch in batches:
        try:
            # Skip running the model if the embeddings already exist
            if all([has_cached_embedding(model_name, filename) for filename in batch]):
                continue

            batch_embeddings, batch_labels, batch_scores = vit.process_images(batch)

            # Save the embeddings
            for emb, pred, score, filename in zip(batch_embeddings, batch_labels, batch_scores, batch):
                emb = emb.astype(np.float32)
                cache_embedding(emb, pred, score, model_name, filename)
        except Exception as e:
            err(f'Error processing {batch}: {e}')


def compute_norm_embedding(model_name: str, images: list, device: str = "cpu"):
    """
    Compute the embedding for a list of images and save them to disk.
    Args:
    :param images:  List of image paths
    :param model_name: Name of the model to use for the embedding generation
    :param device: Device to use for the computation (cpu or cuda:0, cuda:1, etc.)
    Returns:

    """
    # Calculate the mean and standard deviation of the images for normalization
    # This did not work well, but it might be worth revisiting, passing the mean
    # and std to the compute_embedding function
    # mean, std = calc_mean_std(images)
    vit_wrapper = ViTWrapper(device=device, model_name=model_name)

    # If using a GPU, set then skip the parallel CPU processing
    if torch.cuda.is_available():
        compute_embedding_vits(vit_wrapper, images)
    else:
        # Use a pool of processes to speed up the embedding generation 20 images at a time on each process
        num_processes = min(multiprocessing.cpu_count(), len(images) // 20)
        num_processes = max(1, num_processes)
        info(f'Using {num_processes} processes to compute {len(images)} embeddings 20 at a time ...')
        with multiprocessing.Pool(num_processes) as pool:
            args = [(vit_wrapper, images[i:i + 20]) for i in range(0, len(images), 20)]
            pool.starmap(compute_embedding_vits, args)


def calc_mean_std(image_files: list) -> tuple:
    """
    Calculate the mean and standard deviation of a list of images.
    :param image_files: List of absolute image paths
    :return: mean and standard deviation of the images
    """
    mean = np.zeros((224, 224, 3), dtype=np.float32)
    std = np.zeros((224, 224, 3), dtype=np.float32)

    # Loop over all image files in a directory and process them in batches
    for i, file_path in enumerate(image_files):
        # Load the image and compute the mean and standard deviation for each color channel
        img = cv2.imread(file_path)
        # convert from BGR to RGB
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        # Resize the image to 224x224
        img = cv2.resize(img, (224, 224))

        # Compute the mean and standard deviation of the image
        mean += np.mean(img, axis=(0, 1))
        std += np.std(img, axis=(0, 1))

    mean_img = mean.astype(np.uint8) / len(image_files)
    std_img = std.astype(np.uint8) / len(image_files)
    return mean_img, std_img



# Credit: Murat Kirtay, The BioRobotics Inst./SSSA/
# Date: 02/11/2016

import cv2
import numpy as np

def derive_graym(impath):
    """ The intensity value m is calculated as (r+g+b)/3, yet
        grayscale will do same operation!
        opencv uses default formula Y = 0.299 R + 0.587 G + 0.114 B
    """
    # return cv2.imread(impath, cv2.CV_LOAD_IMAGE_GRAYSCALE)
    return cv2.imread(impath, cv2.IMREAD_GRAYSCALE)

def derive_m(img, rimg):
    """ Derive m (intensity) based on paper formula """

    (rw, cl, ch) = img.shape
    for r in range(rw):
        for c in range(cl):
            rimg[r,c] = int(np.sum(img[r,c])/3.0)
            
    return rimg

def derive_saturation(img, rimg):
    """ Derive saturation value for a pixel based on paper formula """

    s_img = np.array(rimg)
    (r, c) = s_img.shape
    for ri in range(r):
        for ci in range(c):
            #opencv ==> b,g,r order
            s1 = np.clip(img[ri,ci][0] + img[ri,ci][2],0,255)
            s2 = 2 * img[ri,ci][1] 
            if  s1 >=  s2:
                s_img[ri,ci] = 1.5*np.clip((img[ri,ci][2] - rimg[ri,ci]),0,255)
            else:
                s_img[ri,ci] = 1.5*np.clip((rimg[ri,ci] - img[ri,ci][0]),0,255)

    return s_img

def check_pixel_specularity(mimg, simg, m_max=0.6, s_max=0.4):
    ''' Check whether a pixel is part of specular region or not'''

    m_max = np.max(mimg) * m_max
    s_max = np.max(simg) * s_max

    (rw, cl) = simg.shape

    spec_mask = np.zeros((rw,cl), dtype=np.uint8)
    for r in range(rw):
        for c in range(cl):
            if mimg[r,c] >= m_max and simg[r,c] <= s_max:
                spec_mask[r,c] = 255
    
    return spec_mask

def enlarge_specularity(spec_mask):
    """ Use sliding window technique to enlarge specularity
        simply move window over the image if specular pixel detected
        mark center pixel is specular
        win_size = 5x5, step_size = 1
    """

    win_size, step_size = (5,5), 1
    enlarged_spec = np.array(spec_mask)
    for r in range(0, spec_mask.shape[0], step_size):
        for c in range(0, spec_mask.shape[1], step_size):
            # yield the current window
            win = spec_mask[r:r + win_size[1], c:c + win_size[0]]
            
            if win.shape[0] == win_size[0] and win.shape[1] == win_size[1]:
                if win[1,1] !=0:
                    enlarged_spec[r:r + win_size[1], c:c + win_size[0]] = 255 * np.ones((5,5), dtype=np.uint8)

    return enlarged_spec



# sdcat, Apache-2.0 license
# Filename: cluster/commands.py
# Description:  Clustering commands

import re
import shutil
import uuid
from datetime import datetime
from pathlib import Path

import click
import ephem
import os
import pandas as pd
import pytz
import torch
from PIL import Image

from sdcat.cluster.utils import clean_bad_images
from sdcat import common_args
from sdcat.config import config as cfg
from sdcat.logger import info, err, warn
from sdcat.cluster.cluster import cluster_vits


@click.command('detections', help='Cluster detections. See cluster --config-ini to override cluster defaults.')
@common_args.config_ini
@common_args.start_image
@common_args.end_image
@common_args.use_tsne
@common_args.skip_visualization
@common_args.alpha
@common_args.cluster_selection_epsilon
@common_args.cluster_selection_method
@common_args.min_cluster_size
@click.option('--det-dir', help='Input folder(s) with raw detection results', multiple=True, required=True)
@click.option('--save-dir', help='Output directory to save clustered detection results', required=True)
@click.option('--device', help='Device to use, e.g. cpu or cuda:0', type=str, default='cpu')
@click.option('--use-vits', help='Set to using the predictions from the vits cluster model', is_flag=True)
def run_cluster_det(det_dir, save_dir, device, use_vits, config_ini, alpha, cluster_selection_epsilon, cluster_selection_method, min_cluster_size, start_image, end_image, use_tsne, skip_visualization):
    config = cfg.Config(config_ini)
    max_area = int(config('cluster', 'max_area'))
    min_area = int(config('cluster', 'min_area'))
    min_saliency = int(config('cluster', 'min_saliency'))
    min_samples = int(config('cluster', 'min_samples'))
    alpha = alpha if alpha else float(config('cluster', 'alpha'))
    min_cluster_size = min_cluster_size if min_cluster_size else int(config('cluster', 'min_cluster_size'))
    cluster_selection_epsilon = cluster_selection_epsilon if cluster_selection_epsilon else float(config('cluster','cluster_selection_epsilon'))
    cluster_selection_method = cluster_selection_method if cluster_selection_method else config('cluster', 'cluster_selection_method')
    remove_corners = config('cluster', 'remove_corners') == 'True'
    remove_bad_images = config('cluster', 'remove_bad_images') == 'True'
    latitude = float(config('cluster', 'latitude'))
    longitude = float(config('cluster', 'longitude'))
    min_score = float(config('cluster', 'min_score'))
    min_similarity = float(config('cluster', 'min_similarity'))
    model = config('cluster', 'model')

    if device != 'cpu':
        num_devices = torch.cuda.device_count()
        info(f'{num_devices} cuda devices available')
        info(f'Using device {device}')
        if 'cuda' in device:
            device_num = device.split(':')[-1]
            info(f'Setting CUDA_VISIBLE_DEVICES to {device_num}')
            torch.cuda.set_device(device)
            os.environ['CUDA_VISIBLE_DEVICES'] = device_num

    save_dir = Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)

    detections = []
    for d in det_dir:
        info(f'Searching in {d}')
        d_path = Path(d)
        if not d_path.exists():
            err(f'Input path {d} does not exist.')
            return

        detections.extend(d_path.rglob('*.csv'))

    # Combine all the detections into a single dataframe
    df = pd.DataFrame()

    crop_path = save_dir / 'crops'
    crop_path.mkdir(parents=True, exist_ok=True)

    for d in detections:
        df_new = pd.read_csv(d, sep=',')

        # concatenate to the df dataframe
        df = pd.concat([df, df_new], ignore_index=True)

    # Remove any duplicate rows; duplicates have the same .x, .y, .xx, .xy,
    df = df.drop_duplicates(subset=['x', 'y', 'xx', 'xy'])

    info(f'Found {len(df)} detections in {det_dir}')

    if len(df) == 0:
        info(f'No detections found in {det_dir}')
        return

    # Check if the image_path column is empty
    if df['image_path'].isnull().values.any():
        err(f'Found {df["image_path"].isnull().sum()} detections with no image_path')
        return

    # Sort the dataframe by image_path to make sure the images are in order for start_image and end_image filtering
    df = df.sort_values(by='image_path')

    # If start_image is set, find the index of the start_image in the list of images
    if start_image:
        start_image = Path(start_image)
        start_image = start_image.resolve()
        start_image = start_image.stem
        start_image_index = df[df['image_path'].str.contains(start_image)].index[0]
        df = df.iloc[start_image_index:]

    # If end_image is set, find the index of the end_image in the list of images
    if end_image:
        end_image = Path(end_image)
        end_image = end_image.resolve()
        end_image = end_image.stem
        end_image_index = df[df['image_path'].str.contains(end_image)].index[0]
        df = df.iloc[:end_image_index]

    # Filter by saliency, area, score or day/night
    size_before = len(df)
    if 'saliency' in df.columns:
        df = df[(df['saliency'] > min_saliency) | (df['saliency'] == -1)]
    if 'area' in df.columns:
        df = df[(df['area'] > min_area) & (df['area'] < max_area)]
    if 'score' in df.columns:
        df = df[(df['score'] > min_score)]
    size_after = len(df)
    info(f'Removed {size_before - size_after} detections outside of area, saliency, or too low scoring')

    # Add in a column for the unique crop name for each detection with a unique id
    # create a unique uuid based on the md5 hash of the box in the row
    df['crop_path'] = df.apply(lambda
                                   row: f"{crop_path}/{uuid.uuid5(uuid.NAMESPACE_DNS, str(row['x']) + str(row['y']) + str(row['xx']) + str(row['xy']))}.png",
                               axis=1)

    # Add in a column for the unique crop name for each detection with a unique id
    df['cluster'] = -1  # -1 is the default value and means that the image is not in a cluster

    # Remove small or large detections before clustering
    size_before = len(df)
    info(f'Searching through {size_before} detections')
    df = df[(df['area'] > min_area) & (df['area'] < max_area)]
    size_after = len(df)
    info(f'Removed {size_before - size_after} detections that were too large or too small')

    def within_1_percent_of_corners(row):
        threshold = 0.01  # 1% threshold

        x, y, xx, yy = row['x'], row['y'], row['xx'], row['xy']

        # Check if any of the coordinates are within 1% of the image corners
        return (
                (0 <= x <= threshold or 1 - threshold <= x <= 1) or
                (0 <= y <= threshold or 1 - threshold <= y <= 1) or
                (0 <= xx <= threshold or 1 - threshold <= xx <= 1) or
                (0 <= yy <= threshold or 1 - threshold <= yy <= 1)
        )

    if remove_corners:
        # Remove any detections that are in any corner of the image
        size_before = len(df)
        df = df[~df.apply(within_1_percent_of_corners, axis=1)]
        size_after = len(df)
        info(f'Removed {size_before - size_after} detections that were in the corners of the image')

    pattern_date1 = re.compile(r'(\d{4})(\d{2})(\d{2})T(\d{2})(\d{2})(\d{2})Z')  # 20161025T184500Z
    pattern_date2 = re.compile(r'(\d{4})(\d{2})(\d{2})T(\d{2})(\d{2})(\d{2})Z\d*mF*')
    pattern_date3 = re.compile(r'(\d{2})(\d{2})(\d{2})T(\d{2})(\d{2})(\d{2})Z')  # 161025T184500Z
    pattern_date4 = re.compile(r'(\d{2})-(\d{2})-(\d{2})T(\d{2})_(\d{2})_(\d{2})-')  # 16-06-06T16_04_54
    pattern_number = re.compile(r'^DSC(\d+)\.JPG')

    # Grab any additional metadata from the image name, e.g. depth, yearday, number, day/night
    depth = {}
    yearday = {}
    number = {}
    day_flag = {}
    observer = ephem.Observer()
    observer.lat = latitude
    observer.lon = longitude

    def is_day(utc_dt):
        observer.date = utc_dt
        sun = ephem.Sun(observer)
        if sun.alt > 0:
            return 1
        return 0

    for index, row in sorted(df.iterrows()):
        image_name = Path(row.image_path).name
        for depth_str in ['50m', '100m', '200m', '300m', '400m', '500m', '299m', '250m', '150m', '199m']:
            if depth_str in image_name:
                depth[index] = int(depth_str.split('m')[0])
        if pattern_date1.search(image_name):
            match = pattern_date1.search(image_name).groups()
            year, month, day, hour, minute, second = map(int, match)
            dt = datetime(year, month, day, hour, minute, second, tzinfo=pytz.utc)
            yearday[index] = int(dt.strftime("%j")) - 1
            day_flag[index] = is_day(dt)
        if pattern_date2.search(image_name):
            match = pattern_date2.search(image_name).groups()
            year, month, day, hour, minute, second = map(int, match)
            dt = datetime(year, month, day, hour, minute, second, tzinfo=pytz.utc)
            yearday[index] = int(dt.strftime("%j")) - 1
            day_flag[index] = is_day(dt)
        if pattern_date3.search(image_name):
            match = pattern_date3.search(image_name).groups()
            year, month, day, hour, minute, second = map(int, match)
            dt = datetime(year, month, day, hour, minute, second, tzinfo=pytz.utc)
            yearday[index] = int(dt.strftime("%j")) - 1
            day_flag[index] = is_day(dt)
        if pattern_date4.search(image_name):
            match = pattern_date4.search(image_name).groups()
            year, month, day, hour, minute, second = map(int, match)
            dt = datetime(year, month, day, hour, minute, second, tzinfo=pytz.utc)
            yearday[index] = int(dt.strftime("%j")) - 1
            day_flag[index] = is_day(dt)
        if pattern_number.search(image_name):
            match = pattern_number.match(image_name)
            numeric = int(match.group(1))
            number[index] = numeric

    # Add the depth, yearday, day, and night columns to the dataframe if they exist
    if len(depth) > 0:
        df['depth'] = depth
        df['depth'] = df['depth'].astype(int)
    if len(yearday) > 0:
        df['yearday'] = yearday
        df['yearday'] = df['yearday'].astype(int)
    if len(number) > 0:
        df['frame'] = number
    if len(day_flag) > 0:
        df['day'] = day_flag
        df['day'] = df['day'].astype(int)

    # Filter by day/night
    # size_before = len(df)
    # df = df[df['day'] == 1]
    # size_after = len(df)
    # info(f'Removed {size_before - size_after} detections that were at night')

    # Replace any NaNs with 0
    df.fillna(0)

    # Print the first 5 rows of the dataframe
    info(df.head(5))

    if len(df) > 0:
        # Replace / with _ in the model name
        model_machine_friendly = model.replace('/', '_')

        # A prefix for the output files to make sure the output is unique for each execution
        prefix = f'{model_machine_friendly}_{datetime.now().strftime("%Y%m%d_%H%M%S")}'

        # Cluster the detections
        df_cluster = cluster_vits(prefix, model, df, save_dir, alpha, cluster_selection_epsilon, cluster_selection_method,
                                  min_similarity, min_cluster_size, min_samples, device, use_tsne=use_tsne,
                                  skip_visualization=skip_visualization, roi=False, use_vits=use_vits,
                                  remove_bad_images=remove_bad_images)

        # Merge the results with the original DataFrame
        df.update(df_cluster)

        # Save the clustered detections to a csv file and a copy of the config.ini file
        df.to_csv(save_dir / f'{prefix}_cluster_detections.csv', index=False, header=True)
        shutil.copy(Path(config_ini), save_dir / f'{prefix}_config.ini')
    else:
        warn(f'No detections found to cluster')
 
@click.command('roi', help='Cluster roi. See cluster --config-ini to override cluster defaults.')
@common_args.config_ini
@common_args.use_tsne
@common_args.skip_visualization
@common_args.alpha
@common_args.cluster_selection_epsilon
@common_args.cluster_selection_method
@common_args.min_cluster_size
@click.option('--roi-dir', help='Input folder(s) with raw ROI images', multiple=True, required=True)
@click.option('--save-dir', help='Output directory to save clustered detection results', required=True)
@click.option('--device', help='Device to use, e.g. cpu or cuda:0', type=str)
@click.option('--use-vits', help='Set to using the predictions from the vits cluster model', is_flag=True)
def run_cluster_roi(roi_dir, save_dir, device, use_vits, config_ini, alpha, cluster_selection_epsilon, cluster_selection_method, min_cluster_size, use_tsne, skip_visualization):
    config = cfg.Config(config_ini)
    min_samples = int(config('cluster', 'min_samples'))
    alpha = alpha if alpha else float(config('cluster', 'alpha'))
    min_cluster_size = min_cluster_size if min_cluster_size else int(config('cluster', 'min_cluster_size'))
    cluster_selection_epsilon = cluster_selection_epsilon if cluster_selection_epsilon else float(config('cluster','cluster_selection_epsilon'))
    cluster_selection_method = cluster_selection_method if cluster_selection_method else config('cluster', 'cluster_selection_method')
    min_similarity = float(config('cluster', 'min_similarity'))
    model = config('cluster', 'model')
    remove_bad_images = config('cluster', 'remove_bad_images') == 'True'

    if device:
        num_devices = torch.cuda.device_count()
        info(f'{num_devices} cuda devices available')
        info(f'Using device {device}')
        if 'cuda' in device:
            device_num = device.split(':')[-1]
            info(f'Setting CUDA_VISIBLE_DEVICES to {device_num}')
            torch.cuda.set_device(device)
            os.environ['CUDA_VISIBLE_DEVICES'] = device_num

    save_dir = Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)

    # Grab all images from the input directories
    supported_extensions = ['.png', '.jpg', '.jpeg', '.JPG', '.JPEG', '.PNG']
    images = []

    for r in roi_dir:
        roi_path = Path(r)
        for ext in supported_extensions:
            images.extend(list(roi_path.rglob(f'*{ext}')))

    # Create a dataframe to store the combined data in an image_path column in sorted order
    df = pd.DataFrame()
    df['image_path'] = images

    # Convert the image_path column to a string
    df['image_path'] = df['image_path'].astype(str)

    info(f'Found {len(df)} detections in {roi_dir}')

    if len(df) == 0:
        info(f'No detections found in {roi_dir}')
        return

    # Sort the dataframe by image_path to make sure the images are in order for start_image and end_image filtering
    df = df.sort_values(by='image_path')

    # Add the image_width and image_height columns to the dataframe
    for index, row in df.iterrows():
        im_size = Image.open(row['image_path']).size
        df.at[index, 'image_width'] = im_size[0]
        df.at[index, 'image_height'] = im_size[1]
    df['image_width'] = df['image_width'].astype(int)
    df['image_height'] = df['image_height'].astype(int)

    # Create a unique crop name for each detection with a unique id
    crop_path = save_dir / 'crops'
    crop_path.mkdir(parents=True, exist_ok=True)
    df['crop_path'] = df.apply(lambda row:
                               f'{crop_path}/{Path(row["image_path"]).stem}.png',
                               axis=1)

    # Add in a column for the unique crop name for each detection with a unique id
    df['cluster'] = -1  # -1 is the default value and means that the image is not in a cluster

    # Replace any NaNs with 0
    df.fillna(0)

    # Print the first 5 rows of the dataframe
    info(df.head(5))

    if len(df) > 0:
        model_machine_friendly = model.replace('/', '_')

        # A prefix for the output files to make sure the output is unique for each execution
        prefix = f'{model_machine_friendly}_{datetime.now().strftime("%Y%m%d_%H%M%S")}'

        # Cluster the detections
        df_cluster = cluster_vits(prefix, model, df, save_dir, alpha, cluster_selection_epsilon, cluster_selection_method,
                                  min_similarity, min_cluster_size, min_samples, device,
                                  use_tsne=use_tsne, use_vits=use_vits,
                                  skip_visualization=skip_visualization,  roi=True,
                                  remove_bad_images=remove_bad_images)

        # Merge the results with the original DataFrame
        df.update(df_cluster)

        # Save the clustered detections to a csv file and a copy of the config.ini file
        df.to_csv(save_dir / f'{prefix}_cluster_detections.csv', index=False, header=True)
        shutil.copy(Path(config_ini), save_dir / f'{prefix}_config.ini')
    else:
        warn(f'No detections found to cluster')



import time

import cv2
import numpy as np

from sdcat.logger import info
from sdcat.reflectance import specularity as spc


def specularity_removal(image: np.ndarray) -> np.ndarray:
    """
    Remove specular highlights from an image. This is done by inpainting the specular highlights.
    Caution: This method is slow for large images. Be sure to resize the image before using this method.
    :param image: color image to remove specular highlights from in RGB format
    :return: cleaned image
    """
    start_time = time.time()
    # Radius for inpainting should be smaller for smaller images, but max out at 30
    radius = min(int(min(image.shape[:2]) * .01), 30)
    info(f'Specularity removal using radius {radius} for inpainting')

    img_color = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    img_gray = cv2.cvtColor(img_color, cv2.COLOR_BGR2GRAY)

    r_img = np.array(img_gray)
    rimg = spc.derive_m(img_color, r_img)
    # TODO: debug overflows by changing the dtype to uint16 and scaling the values
    s_img = spc.derive_saturation(img_color, rimg)
    spec_mask = spc.check_pixel_specularity(rimg, s_img)
    enlarged_spec = spc.enlarge_specularity(spec_mask)
    # Use inpaint methods to remove specularity
    telea = cv2.inpaint(img_color, enlarged_spec, radius, cv2.INPAINT_TELEA)

    end_time = time.time()
    info(f"Specularity removal took {end_time - start_time:.2f} seconds")
    clean_img = cv2.cvtColor(telea, cv2.COLOR_BGR2RGB)
    return clean_img



# sdcat, Apache-2.0 license
# Filename: sdcat/logger/__init__.py
# Description:  Logs to both a file and the console. The file is named with the current date.

import logging
import os
from pathlib import Path
from datetime import datetime as dt, timezone

LOGGER_NAME = "sdcat"
DEBUG = True


class _Singleton(type):
    """A metaclass that creates a Singleton base class when called."""

    _instances = {}  # type: ignore

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Singleton(_Singleton("SingletonMeta", (object,), {})):
    pass


class CustomLogger(Singleton):
    logger = None
    output_path = Path.home() / "sdcat" / "logs"

    def __init__(self, output_path: Path = Path.cwd(), output_prefix: str = LOGGER_NAME):
        """
        Initialize the logger
        """
        self.logger = logging.getLogger(LOGGER_NAME)
        self.logger.setLevel(logging.DEBUG)
        self.output_path = output_path
        self.output_path.mkdir(parents=True, exist_ok=True)
        formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")

        # default log file date to today
        now = dt.now(timezone.utc)

        # log to file
        self.log_filename = output_path / f"{output_prefix}_{now:%Y%m%d}.log"
        handler = logging.FileHandler(self.log_filename, mode="w")
        handler.setFormatter(formatter)
        handler.setLevel(logging.INFO)
        self.logger.addHandler(handler)

        # also log to console
        console = logging.StreamHandler()
        console.setLevel(logging.DEBUG)
        console.setFormatter(formatter)
        self.logger.addHandler(console)

        self.logger.info(f"Logging to {self.log_filename}")

    def loggers(self) -> logging.Logger:
        return self.logger


def create_logger_file(prefix: str = "sdcat"):
    """
    Create a logger file
    :param log_path: Path to the log file
    """
    ENVIRONMENT = str(os.getenv("ENVIRONMENT"))
    if ENVIRONMENT and ENVIRONMENT.upper() == "TESTING":
        log_path = Path("logs")
    else:
        log_path = Path.home() / "sdcat" / "logs"
        # Check if can write to the log path, and if not revert to system temp
        try:
            log_path.mkdir(parents=True, exist_ok=True)
            test_file = log_path / "test.txt"
            test_file.touch()
            test_file.unlink()
        except PermissionError:
            import tempfile

            temp_dir = tempfile.gettempdir()
            log_path = Path(temp_dir) / "sdcat" / "logs"

    # create the log directory if it doesn't exist
    log_path.mkdir(parents=True, exist_ok=True)
    return CustomLogger(log_path, prefix)


def custom_logger() -> logging.Logger:
    """
    Get the logger
    """
    return logging.getLogger(LOGGER_NAME)


def err(s: str):
    custom_logger().error(s)


def info(s: str):
    custom_logger().info(s)


def debug(s: str):
    custom_logger().debug(s)


def warn(s: str):
    custom_logger().warning(s)


def exception(s: str):
    custom_logger().exception(s)


def critical(s: str):
    custom_logger().critical(s)



# sdcat, Apache-2.0 license
# Filename: sdcat/config/config.py
# Description:   Configuration helper to setup defaults for sdcat

from configparser import ConfigParser
import os
from sdcat.logger import info

default_config_ini = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'config.ini')


class Config:

    def __init__(self, path: str = None, quiet: bool = False):
        """
        Read the .ini file and parse it
        """
        self.parser = ConfigParser()
        if path:
            self.path = path
        else:
            self.path = default_config_ini

        if not os.path.isfile(self.path):
            info(f'Bad path to {self.path}. Is your {self.path} missing?')
            raise Exception(f'Bad path to {self.path}. Is your {self.path} missing?')

        self.parser.read(self.path)
        lines = open(self.path).readlines()
        if not quiet:
            info(f"=============== Config file {self.path} =================")
            for l in lines:
                info(l.strip())

            if not path:
                info(f"============ You can override these settings by creating a customconfig.ini file and pass that "
                     f"in with --config=customconfig.ini =====")

    def __call__(self, *args, **kwargs):
        assert len(args) == 2
        return self.parser.get(args[0], args[1])

    def save(self, *args, **kwargs):
        assert len(args) == 3
        self.parser.set(section=args[0], option=args[1], value=args[2])
        with open(self.path, 'w') as fp:
            self.parser.write(fp)



import hashlib
import multiprocessing
import shutil
from pathlib import Path

import click
from tqdm import tqdm
import cv2
import pandas as pd
import torch
from sahi.postprocess.combine import nms

from sdcat import common_args
from sdcat.config import config as cfg
from sdcat.config.config import default_config_ini
from sdcat.detect.sahi_detector import run_sahi_detect_bulk, run_sahi_detect
from sdcat.detect.saliency_detector import run_saliency_detect, run_saliency_detect_bulk
from sdcat.logger import exception, info, warn, create_logger_file

default_model = 'MBARI/megamidwater'


@click.command('detect',
               help=f'Detect objects in images.  By default, runs both SAHI and saliency blob detection and combines using NMS. '
                    f'Use --skip-sahi or --skip-saliency to exclude. '
                    f'See --config-ini to override detection defaults in {default_config_ini}.')
@common_args.config_ini
@common_args.start_image
@common_args.end_image
@click.option('--show', is_flag=True, help='Show algorithm steps.')
@click.option('--image-dir', required=True, help='Directory with images to run sliced detection.')
@click.option('--save-dir', required=True, help='Save detections to this directory.')
@click.option('--device', default='cpu', help='Device to use, e.g. cpu or cuda:0')
@click.option('--spec-remove', is_flag=True, help='Run specularity removal algorithm on the images before processing. '
                                                  '**CAUTION**this is slow. Set --scale-percent to < 100 to speed-up')
@click.option('--skip-sahi', is_flag=True, help='Skip sahi detection.')
@click.option('--skip-saliency', is_flag=True, help='Skip saliency detection.')
@click.option('--conf', default=0.1, help='Confidence threshold.')
@click.option('--scale-percent', default=80, help='Scaling factor to rescale the images before processing.')
@click.option('--model', default=default_model, help=f'Model to use. Defaults to {default_model}')
@click.option('--slice-size-width', help='Slice width size, leave blank for auto slicing')
@click.option('--slice-size-height', help='Slice height size, leave blank for auto slicing')
@click.option('--postprocess-match-metric', default='IOS', help='Postprocess match metric for NMS. postprocess_match_metric IOU for intersection over union, IOS for intersection over smaller area.')
@click.option('--overlap-width-ratio', default=0.4, help='Overlap width ratio for NMS')
@click.option('--overlap-height-ratio', default=0.4, help='Overlap height ratio for NMS')
@click.option('--clahe', is_flag=True, help='Run the CLAHE algorithm to contrast enhance before detection useful images with non-uniform lighting')

def run_detect(show: bool, image_dir: str, save_dir: str, model: str,
               slice_size_width: int, slice_size_height: int, scale_percent: int,
                postprocess_match_metric: str, overlap_width_ratio: float, overlap_height_ratio: float,
               device: str, conf: float, skip_sahi: bool, skip_saliency: bool, spec_remove: bool,
               config_ini: str, clahe: bool, start_image: str, end_image: str):
    config = cfg.Config(config_ini)
    clahe = clahe if clahe else config('detect', 'clahe') == 'True'
    block_size = int(config('detect', 'block_size'))
    min_std = float(config('detect', 'min_std'))
    max_area = int(config('detect', 'max_area'))
    min_area = int(config('detect', 'min_area'))
    min_saliency = int(config('detect', 'min_saliency'))
    class_agnostic = config('detect', 'class_agnostic') == 'True'
    allowable_classes = []
    if not class_agnostic:
        allowable_classes = config('detect', 'allowable_classes')
        if len(allowable_classes) > 0:
            allowable_classes = allowable_classes.split(',')

    create_logger_file('detect')

    if not skip_sahi:
        from sahi import AutoDetectionModel
        if model == 'yolov8s':
            detection_model = AutoDetectionModel.from_pretrained(
                model_type='yolov8',
                model_path='ultralyticsplus/yolov8s',
                confidence_threshold=conf,
                device=device,
            )
        elif model == 'yolov8x':
            detection_model = AutoDetectionModel.from_pretrained(
                model_type='yolov8',
                model_path='yolov8x.pt',
                confidence_threshold=conf,
                device=device,
            )
        elif model == 'hustvl/yolos-small':
            model_path = 'hustvl/yolos-small'
            detection_model = AutoDetectionModel.from_pretrained(
                model_type='huggingface',
                model_path=model_path,
                config_path=model_path,
                confidence_threshold=conf,
                device=device,
            )
        elif model == 'hustvl/yolos-tiny':
            model_path = 'hustvl/yolos-tiny'
            detection_model = AutoDetectionModel.from_pretrained(
                model_type='huggingface',
                model_path=model_path,
                config_path=model_path,
                confidence_threshold=conf,
                device=device,
            )
        elif model == 'MBARI/megamidwater':
            # Download model path
            from huggingface_hub import hf_hub_download
            model_path = hf_hub_download(repo_id="MBARI-org/megamidwater", filename="best.pt")
            detection_model = AutoDetectionModel.from_pretrained(
                model_type='yolov5',
                model_path=model_path,
                config_path=model_path,
                confidence_threshold=conf,
                device=device,
            )
        elif model == 'MBARI/uav-yolov5-30k':
            # Download model path
            from huggingface_hub import hf_hub_download
            model_path = hf_hub_download(repo_id="MBARI-org/yolov5x6-uav-30k", filename="yolov5x6-uav-30k.pt")
            detection_model = AutoDetectionModel.from_pretrained(
                model_type='yolov5',
                model_path=model_path,
                config_path=model_path,
                confidence_threshold=conf,
                device=device,
            )
        elif model == 'MBARI/uav-yolov5-18k':
            # Download model path
            from huggingface_hub import hf_hub_download
            model_path = hf_hub_download(repo_id="MBARI-org/yolov5-uav-18k", filename="yolov5x6-uav-18k.pt")
            detection_model = AutoDetectionModel.from_pretrained(
                model_type='yolov5',
                model_path=model_path,
                config_path=model_path,
                confidence_threshold=conf,
                device=device,
            )
        elif model == 'MBARI/yolov5x6-uavs-oneclass':
            # Download model path
            from huggingface_hub import hf_hub_download
            model_path = hf_hub_download(repo_id="MBARI-org/yolov5x6-uavs-oneclass", filename="best_uavs_oneclass.pt")

            detection_model = AutoDetectionModel.from_pretrained(
                model_type='yolov5',
                model_path=model_path,
                config_path=model_path,
                confidence_threshold=conf,
                device=device,
            )
        elif model == 'FathomNet/MBARI-315k-yolov5':
            # Download model path
            from huggingface_hub import hf_hub_download
            model_path = hf_hub_download(repo_id="FathomNet/MBARI-315k-yolov5", filename="mbari_315k_yolov5.pt")
            detection_model = AutoDetectionModel.from_pretrained(
                model_type='yolov5',
                model_path=model_path,
                config_path=model_path,
                confidence_threshold=conf,
                device=device,
            )
        else:
            exception(f'Unknown model: {model}')
            return

    images_path = Path(image_dir)
    base_path = Path(save_dir) / model

    # clean-up any previous results
    if base_path.exists():
        info(f'Removing {base_path}')
        shutil.rmtree(base_path)

    if not skip_sahi:
        save_path_base = Path(save_dir) / model
    else:
        save_path_base = Path(save_dir)

    save_path_det_raw = save_path_base / 'det_raw' / 'csv'
    save_path_det_filtered = save_path_base / 'det_filtered' / 'csv'
    save_path_viz = save_path_base / 'vizresults'

    save_path_det_raw.mkdir(parents=True, exist_ok=True)
    save_path_det_filtered.mkdir(parents=True, exist_ok=True)
    save_path_viz.mkdir(parents=True, exist_ok=True)

    # Run on all images recursively. Sort the dataframe by image_path to make sure the images are in order for start_image and end_image filtering
    # Find all valid images
    images = [file for file in sorted(images_path.rglob('*'))
              if file.as_posix().endswith(('jpeg', 'png', 'jpg', 'JPEG', 'PNG', 'JPG', 'tif', 'tiff'))]

    # If start_image is set, find the index of the start_image in the list of images
    if start_image:
        start_image = Path(start_image)
        start_image_index = next((i for i, image in enumerate(images) if start_image.name in image.name), None)
        if start_image_index is None:
            warn(f'Start image {start_image} not found in images')
            return
        images = images[start_image_index:]

    # If end_image is set, find the index of the end_image in the list of images
    if end_image:
        end_image = Path(end_image)
        end_image_index = next((i for i, image in enumerate(images) if end_image.name in image.name), None)
        if end_image_index is None:
            warn(f'End image {end_image} not found in images')
            return
        images = images[:end_image_index + 1]

    num_images = len(images)
    info(f'Found {num_images} images in {images_path}')

    if num_images == 0:
        return

    if not skip_saliency:
        # run_saliency_detect(spec_remove, scale_percent, images[0].as_posix(), (save_path_det_raw / f'{images[0].stem}.csv').as_posix(), clahe=clahe, show=True)
        # Do the work in parallel to speed up the processing on multicore machines
        num_processes = multiprocessing.cpu_count()
        num_processes = min(num_processes, num_images)
        info(f'Using {num_processes} processes to compute {num_images} images 10 at a time ...')
        # # Run multiple processes in parallel num_cpu images at a time
        with multiprocessing.Pool(num_processes) as pool:
            args = [(spec_remove,
                     scale_percent,
                     images[i:i + 1],
                     save_path_det_raw,
                     min_std,
                     block_size,
                     clahe,
                     show)
                    for i in range(0, num_images, 1)]
            pool.starmap(run_saliency_detect_bulk, args)
            pool.close()
    if device == 'cpu':
        with multiprocessing.Pool(num_processes) as pool:
            if not skip_sahi:
                # Run sahi detection on each image
                args = [(scale_percent,
                         slice_size_width,
                         slice_size_height,
                         [image],
                         save_path_det_raw,
                         detection_model,
                         postprocess_match_metric,
                         overlap_width_ratio,
                         overlap_height_ratio,
                         allowable_classes,
                         class_agnostic)
                        for image in images]
                pool.starmap(run_sahi_detect_bulk, args)
                pool.close()
    else:
        for f in tqdm(images):
            if not skip_saliency:
                run_saliency_detect(spec_remove,
                                    scale_percent,
                                    f.as_posix(),
                                    (save_path_det_raw / f'{f.stem}.csv').as_posix(),
                                    None,
                                    min_std,
                                    block_size,
                                    clahe,
                                    show)
            if not skip_sahi:
                run_sahi_detect(scale_percent,
                                slice_size_width,
                                slice_size_height,
                                f,
                                (save_path_det_raw / f'{f.stem}.sahi.csv'),
                                detection_model,
                                postprocess_match_metric,
                                overlap_width_ratio,
                                overlap_height_ratio,
                                allowable_classes,
                                class_agnostic)

    # Combine all the detections into a single dataframe per image
    for f in tqdm(images):
        # Get the width and height of the image
        img_color = cv2.imread(f.as_posix())
        height, width = img_color.shape[:2]

        # Dataframe to store all the csv files for this image
        df_combined = pd.DataFrame()

        # Path to save final csv file with the detections
        pred_out_csv = save_path_det_filtered / f'{f.stem}.csv'

        for csv_file in save_path_det_raw.rglob(f'{f.stem}*csv'):
            # Read in the csv file and add it to the combined dataframe
            df = pd.read_csv(csv_file, sep=',')
            df_combined = pd.concat([df_combined, df])

        info(f'Found {len(df_combined)} detections in {f}')

        if len(df_combined) == 0:
            warn(f'No detections found in {f}')
            continue

        size_before = len(df_combined)
        df_combined = df_combined[(df_combined['area'] > min_area) & (df_combined['area'] < max_area)]
        size_after = len(df_combined)
        info(f'Removed {size_before - size_after} detections that were too large or too small')

        size_before = len(df_combined)
        # allow negative saliency which means the saliency was not included - this is the case if running SAHI algorithm only and no saliency blob detect
        df_combined = df_combined[(df_combined['saliency'] > min_saliency) | (df_combined['saliency'] == -1)]
        size_after = len(df_combined)
        info(f'Removed {size_before - size_after} detections that were low saliency')

        if len(df_combined) == 0:
            warn(f'No detections found in {f}')
            continue

        # Convert the x, y, xx, xy, score columns to a list of Tensors that is Shape: [num_boxes,5].
        pred_list = df_combined[['x', 'y', 'xx', 'xy', 'score']].values.tolist()
        pred_list = torch.tensor(pred_list)  # convert the list of Tensors to a list of Tensors

        # Clean up the predictions using NMS
        info(f'Running NMS on {len(df_combined)} predictions')
        nms_pred_idx = nms(pred_list, 'IOU', 0.1)
        info(f'{len(nms_pred_idx)} predictions found after NMS')

        # Filter the original DataFrame based on the indices kept by NMS and keep the saliency and area columns
        df_final = df_combined.iloc[nms_pred_idx].reset_index(drop=True)
        df_final['saliency'] = df_combined['saliency'].iloc[nms_pred_idx].reset_index(drop=True)
        df_final['area'] = df_combined['area'].iloc[nms_pred_idx].reset_index(drop=True)

        # Plot boxes on the input frame
        pred_list = df_final[['x', 'y', 'xx', 'xy', 'score', 'class']].values.tolist()
        for p in pred_list:
            if class_agnostic:
                img_color = cv2.rectangle(img_color,
                                          (int(p[0]), int(p[1])),
                                          (int(p[2]), int(p[3])), (81, 12, 51), 3)
                img_color = cv2.putText(img_color, f'{p[5]} {p[4]:.2f}', (int(p[0]), int(p[1])),
                                        cv2.FONT_HERSHEY_SIMPLEX, 1, (255, 255, 255), 2, cv2.LINE_AA)
            else:
                # Color based on the first 5 letters of the label so that the same label always gets the same color
                md5_hash = hashlib.md5(p[5].encode())
                hex_color = md5_hash.hexdigest()[:6]
                r, g, b = (int(hex_color[:2], 16), int(hex_color[2:4], 16), int(hex_color[4:], 16))
                color = (r % 256, g % 256, b % 256)
                img_color = cv2.rectangle(img_color,
                                          (int(p[0]), int(p[1])),
                                          (int(p[2]), int(p[3])), color, 3)
                img_color = cv2.putText(img_color, p[5], (int(p[0]), int(p[1])), cv2.FONT_HERSHEY_SIMPLEX, 1, color, 2,
                                        cv2.LINE_AA)
        info(f'Saving visualization to {save_path_viz / f"{f.stem}.jpg"}')
        cv2.imwrite(f"{save_path_viz / f'{f.stem}.jpg'}", img_color)

        df_final['cluster'] = -1
        df_final['image_path'] = f.as_posix()
        df_final['image_width'] = width
        df_final['image_height'] = height

        # Normalize the predictions to 0-1.
        df_final['x'] = df_final['x'] / width
        df_final['y'] = df_final['y'] / height
        df_final['xx'] = df_final['xx'] / width
        df_final['xy'] = df_final['xy'] / height
        df_final['w'] = (df_final['xx'] - df_final['x'])
        df_final['h'] = (df_final['xy'] - df_final['y'])

        # Save DataFrame to CSV file including image_width and image_height columns
        df_final.to_csv(pred_out_csv.as_posix(), index=False, header=True)

        info(f'Found {len(pred_list)} total localizations in {f} with {len(df_combined)} after NMS')
        info(f'Slice width: {slice_size_width} height: {slice_size_height}')

        save_stats = save_path_base / 'stats.txt'
        with open(save_stats, 'w') as sf:
            sf.write(f"Statistics for {f}:\n")
            sf.write("----------------------------------\n")
            sf.write(f"Total number of bounding boxes: {df_combined.shape[0]}\n")
            sf.write(
                f"Total number of images with (bounding box) detections found: {df_combined['image_path'].nunique()}\n")
            sf.write(
                f"Average number of bounding boxes per image: {df_combined.shape[0] / df_combined['image_path'].nunique()}\n")
            sf.write(f"Average width of bounding boxes: {df_combined['w'].mean() * width}\n")
            sf.write(f"Average height of bounding boxes: {df_combined['h'].mean() * height}\n")
            sf.write(f"Average area of bounding boxes: {df_combined['area'].mean()}\n")
            sf.write(f"Average score of bounding boxes: {df_combined['score'].mean()}\n")
            sf.write(f"Average saliency of bounding boxes: {df_combined['saliency'].mean()}\n")



# sdcat, Apache-2.0 license
# Filename: common_args.py
# Description: Common arguments for processing commands

import click

from sdcat.config.config import default_config_ini

# Common arguments for processing commands
config_ini = click.option('--config-ini',
                          type=str,
                          default=default_config_ini,
                          help=f'Path to config file to override. Defaults are in {default_config_ini}. Copy to your own custom.ini file to override')

start_image = click.option('--start-image',
                            type=str,
                            help='Start image name')

end_image = click.option('--end-image',
                            type=str,
                            help='End image name')

alpha = click.option('--alpha',
                     type=float,
                     help='Alpha is a parameter that controls the linkage. See https://hdbscan.readthedocs.io/en/latest/parameter_selection.html. '
                          'Default is 0.92. Increase for less conservative clustering, e.g. 1.0')

cluster_selection_epsilon = click.option('--cluster-selection-epsilon',
                                         type=float,
                                         help='Epsilon is a parameter that controls the linkage. '
                                              'Default is 0. Increase for less conservative clustering')

cluster_selection_method = click.option('--cluster-selection-method',
                                        type=str,
                                        default='leaf',
                                        help='Method for selecting the optimal number of clusters. '
                                             'Default is leaf. Options are leaf, eom, and dill')

min_cluster_size = click.option('--min-cluster-size',
                                type=int,
                                help='The minimum number of samples in a group for that group to be considered a cluster. '
                                     'Default is 2. Increase for less conservative clustering, e.g. 5, 15')

use_tsne = click.option('--use-tsne',
                        is_flag=True,
                        help='Use t-SNE for dimensionality reduction. Default is False')

skip_visualization = click.option('--skip-visualization',
                                    is_flag=True,
                                    help='Skip visualization. Default is False')


# sdcat, Apache-2.0 license
# Filename: sdcat/ml/sahi.py
# Description:  SAHI detection models
import numpy as np
from pathlib import Path

import cv2
import pandas as pd
import torch
from sahi.predict import get_sliced_prediction

from sdcat.logger import info, exception
from sdcat.cluster.utils import rescale
from skimage.morphology import disk
from skimage.filters import rank


def local_equalize(image: np.ndarray, size: int = 3) -> np.ndarray:
    img = cv2.convertScaleAbs(image)  # convert to 8-bit
    footprint = disk(size)
    return rank.equalize(img, footprint=footprint)


def run_sahi_detect_bulk(scale_percent: int,
                         slice_width: int,
                         slice_height: int,
                         images: list,
                         csv_out_path: Path,
                         detection_model,
                         postprocess_match_metric: str,
                         overlap_width_ratio: float,
                         overlap_height_ratio: float,
                         allowable_classes: list = None,
                         class_agnostic: bool = False):
    info(f'Processing {len(images)} images')
    if not csv_out_path.is_dir():
        csv_out_path.mkdir(parents=True, exist_ok=True)
    for f in images:
        run_sahi_detect(scale_percent,
                        slice_width,
                        slice_height,
                        f,
                        (csv_out_path / f'{f.stem}.sahi.csv'),
                        detection_model,
                        postprocess_match_metric,
                        overlap_width_ratio,
                        overlap_height_ratio,
                        allowable_classes,
                        class_agnostic)


def run_sahi_detect(scale_percent: int,
                    slice_width: int,
                    slice_height: int,
                    image_path: Path,
                    csv_out_path: Path,
                    detection_model,
                    postprocess_match_metric: str = 'IOS',
                    overlap_width_ratio: float = 0.4,
                    overlap_height_ratio: float = 0.4,
                    allowable_classes: list = None,
                    class_agnostic: bool = False) -> pd.DataFrame:
    """
    Run the sahi detection model on an image
    :param scale_percent: percent to scale the image
    :param slice_width: slice size width
    :param slice_height: slice size height
    :param image_path: path to the image
    :param csv_out_path: output path for the detections
    :param image_color: color image
    :param detection_model: detection model
    :param postprocess_match_metric 'IOU' for intersection over union, 'IOS' for intersection over smaller area.
    :param overlap_width_ratio: overlap width ratio - amount of overlap between slices
    :param overlap_height_ratio: overlap height ratio - amount of overlap between slices
    :param allowable_classes: list of allowable classes
    :param class_agnostic: True if class agnostic
    :return: dataframe of detections
    """
    df = pd.DataFrame()
    # Abort if the image_path is not a file
    if not image_path.is_file():
        exception(f'Image path {image_path} is not a file')
        return df
    info(f'Processing {image_path}')
    img_color = cv2.imread(image_path.as_posix())
    img_color_rescaled = rescale(img_color, scale_percent=scale_percent)

    # Calculate the slice size
    if slice_width and slice_height:
        info(f'Using slice size width: {slice_width} and height: {slice_height}')
        slice_height = int(slice_height)
        slice_width = int(slice_width)
    else:
        # Let the auto-slice do its thing
        slice_width = None
        slice_height = None

    try:
        result = get_sliced_prediction(
            img_color_rescaled,
            detection_model,
            slice_height=slice_height,
            slice_width=slice_width,
            overlap_height_ratio=overlap_height_ratio,
            overlap_width_ratio=overlap_width_ratio,
            postprocess_type="NMM",
            postprocess_match_metric=postprocess_match_metric,
            perform_standard_pred=True,
            # Perform a standard prediction on top of sliced predictions to increase large object detection
            postprocess_class_agnostic=class_agnostic,  # If True, postprocess will ignore category ids.
        )

        predictions = []

        for obj in result.object_prediction_list:
            info(obj)
            if class_agnostic:
                obj.category.name = 'Unknown'
                obj.category.id = 0
                predictions.append(obj)
            if allowable_classes:
                if obj.category.name in allowable_classes:
                    predictions.append(obj)
            else:
                predictions.append(obj)

        if len(predictions) > 0:

            # Calculate the scale factor to rescale the detections back to the original image size
            scale_width, scale_height = img_color.shape[0] / img_color_rescaled.shape[0], img_color.shape[1] / \
                                        img_color_rescaled.shape[1]

            for p in predictions:
                box_width = p.bbox.maxx - p.bbox.minx
                box_height = p.bbox.maxy - p.bbox.miny
                area = box_width * box_height  # an approximation of the area
                score = p.score.value
                class_label = p.category.name
                # get the scores from the tensor
                if isinstance(score, torch.Tensor):
                    score = score.item()
                df = pd.concat([df, pd.DataFrame({
                    'image_path': image_path.as_posix(),
                    'class': class_label,
                    'score': score,
                    'area': area,
                    'saliency': -1,  # not computed - just a placeholder
                    'x': p.bbox.minx,
                    'y': p.bbox.miny,
                    'xx': p.bbox.maxx,
                    'xy': p.bbox.maxy,
                    'w': box_width,
                    'h': box_height},
                    index=[0])])

            # Remove duplicates, a duplicate has the same class and h and w
            df = df.drop_duplicates(subset=['class', 'h', 'w'])

            # Scale the detections back to the original image size
            df[['x', 'xx', 'w']] *= scale_width
            df[['y', 'xy', 'h']] *= scale_height

            # Write out the detections
            df.to_csv(csv_out_path.as_posix(), index=False)
    except Exception as e:
        exception(f"Error processing {image_path}: {e}")

    return df
